//	===============================================
//	RGB Led-Leiste mit 10 LEDs
//	===============================================
//  RGB_LED_Main.c
//	Version 2.00 (21.06.2007)
//
//	System arbeitet mit 8,0 MHz
//
//	*** Timer ***:
//	Timer 0 Peak hold Timer fr VU-Meter Modus
//	Timer 1 arbeitet mit Voteiler 64 und TopWert 6250 => genaues 0,05s Timing
//	Timer 2 zustndig fr LED Helligkeit mit Vorteiler 8 =>	125kHz LED frequenz
//
//	USART asynchron mit 38,4k (UBRR = 12)
//
//	*** Ports ***:
//	PortA-Belegung:
//					LED 1-8 Grn
//			
//	PortB-Belegung:
//					LED 1-8 Rot
//			
//	PortC-Belegung:
//					LED 1-8 Blau
//			
//	PortD-Belegung:		
//					Bit 0 - UART RXD
//					Bit 1 - UART TXD
//					Bit 2+3 - LED 9+10 Blau
//					Bit 4+5 - LED 9+10 Grn
//					Bit 6+7 - LED 9+10 Rot

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <stdint.h>
#include "RGB_LED.h"
#include "RGB_LED_Main.h"
#include "RGB_LED_Prog.h"

/********************************************************
*				Variablen Definitionen					*
********************************************************/

register unsigned char ProgCtrlReg asm("r3");	// -> Festes Register
register unsigned char ModeOptions asm("r4");

volatile uint16_t ProgTime;
volatile uint16_t ProgTime_Max;
volatile uint16_t NextEventTime;

volatile uint8_t EventAnzahl;
volatile uint8_t NextEvent;

// EEPROM Variable zum spechern des letzten Programmzustands
uint8_t eeLastProgStat EEMEM;
uint8_t tempLastProgStat;

// Programm zwischenspeicher
uint8_t tempProgDaten;
uint8_t tempHelligkeit;
uint8_t *ProgPointer;

uint8_t CurCom;
uint8_t LED_HellPosOffset;
uint8_t i;

// Programm Daten
/* LED_Helligkeits_Cycle:
*  2 Cycle (Front/Back) * LED_HellCycleLeng (Zyklus Dauer) * 4 Ports
*/
uint8_t LED_Helligkeits_Cycle[2][LED_HellCycleLeng][4];
uint8_t VU_Current_Color[4];
uint8_t DM_New_LEDSet[30];


/********************************************************
*				main() - Hauptprogramm					*
********************************************************/

int main (void) {
	// Inititalisierung
	// == Init UART ==
	UBRRH = 0;
	UBRRL = 12;
	UCSRB = (1<<TXEN)|(1<<RXEN)|(1<<RXCIE);	// Enable TX, RX and RX Interrupt
	UCSRC |= (1<<URSEL)|(3<<UCSZ0); 			// Asynchron 8N1

	// == Set I/O Ports ==
	DDRA = 0b11111111; // *** LED outputs LED 1-8 ***
	DDRB = 0b11111111; // *** LED outputs LED 1-8 ***
	DDRC = 0b11111111; // *** LED outputs LED 1-8 ***
	DDRD = 0b11111100;	// *** LED outputs LED 9+10 ***

	PORTA = 0xFF;
	PORTB = 0xFF;
	PORTC = 0xFF;
	PORTD = 0xFC;

	// == Prepare Timer 0 ==
	TCNT0 = 0x00;
	TCCR0 = 0x00;			// Timer nicht gestartet

	// == Prepare Timer 1 ==
	OCR1A = Timer1_Top;		// TopWert 6250 => genaues 0,05s Timing
	TCCR1A = 0x00;
	TCCR1B = 0x00;			// Timer nicht gestartet

	//TCCR1B = (1 << WGM12) | (3 << CS10);		// CTC Mode (OCR1A = Top) vorteiler 64
	TCNT1 = 0x0000;			// Zero counter

	// == Prepare Timer 2 ==
	TCNT2 = 0x00;
	TCCR2 = 0x00;			// Timer nicht gestartet

	//== Set Interupts ==
	TIMSK = (1 << TOIE0) | (1 << TOIE2) | (1 << OCIE1A);	//Timer 0 + 2 overflow, Timer 1 compare Match A interupt enable
	sei();		//Enable global Interrup

	// ** Init Program Mode / check last Mode
	tempLastProgStat = eeprom_read_byte(&eeLastProgStat);
	if (tempLastProgStat == 0x00) {
		ProgCtrlReg = (1 << PC_Goto_PMode);	// default: Program Modus
	}
	else {
		ProgCtrlReg = (tempLastProgStat << 1); // letzten Modus
	}

	//	=============
	//	Hauptschleife
	//	=============

	while(1) {
		// **** Event Handler ****
		//## Bootloader Starten ##
		if ((ProgCtrlReg & (1 << PC_Start_Bootload)) == (1 << PC_Start_Bootload)) {
			//bootload_init();
		}
		//## Wechseln zu VU-Meter Modus (Winamp) ##
		if ((ProgCtrlReg & (1 << PC_Goto_VUMode)) == (1 << PC_Goto_VUMode)) {
			VUMode_Init();
		}
		//## Wechseln zu Programm Modus (LED sequenz abspielen) ##
		if ((ProgCtrlReg & (1 << PC_Goto_PMode)) == (1 << PC_Goto_PMode)) {
			PMode_Init();
		}
		//## Wechseln zu Direkt Modus (LEDs werden direkt von PC geschaltet) ##
		if ((ProgCtrlReg & (1 << PC_Goto_DMode)) == (1 << PC_Goto_DMode)) {
			DMode_Init();
		}
		//## Vorbereiten der nchsten LED konstellation (Nur PMode) ##
		if ((ProgCtrlReg & (1 << PC_PM_NextLEDSet)) == (1 << PC_PM_NextLEDSet)) {
			PMode_PrepareLedSet();
		}
		//## Vorbereiten der neuen LED konstellation (Nur DMode) ##
		if ((ProgCtrlReg & (1 << PC_DM_NewLEDSet)) == (1 << PC_DM_NewLEDSet)) {
			DMode_NewLedSet();
		}


		//## Nichts zu tun -> Loop
	}
	/* wird nie erreicht */
	return 0;
}

/********************************************************
*			VU-Meter Modus Initialisieren				*
********************************************************/

void VUMode_Init() {

}

/********************************************************
*			Program Modus Initialisieren				*
********************************************************/

void PMode_Init() {
	// ** Neuen Modus in ProgCtrlReg eintragen
	ProgCtrlReg &= (1 << PC_Start_Bootload);
	ModeOptions = (1 << MC_PMode);
	tempLastProgStat = eeprom_read_byte(&eeLastProgStat);
	if (tempLastProgStat != (ModeOptions & (0b00000111))) {
		eeprom_write_byte(&eeLastProgStat, ModeOptions & (0b00000111));
	}

	LED_HellPosOffset = 0x00;
	CurCom = 0x00;

	// ** Timer Stoppen
	TCCR1B = 0x00;
	TCCR2 = 0x00;
	TCCR0 = 0x00;

	// ** Aktuelle Zeit und auf Null
	ProgTime = 0x00;

	// ** Program Optionen laden
	ProgTime_Max = pgm_read_word(&ProgramOptions);
	EventAnzahl = pgm_read_byte(&ProgramOptions[1]);

	// ** Beginne Laden von erstem Event
	NextEvent = 0x00;
	PMode_PrepareLedSet();

	// ** Prfen ob erstes Event bei t0
	if (NextEventTime == 0x00) {
		// ** Aktuelle Event Helligkeit einstellen und Befehl zum Event nachladen
		ModeOptions |= (1 << MC_MemPointer);
		ProgCtrlReg |= (1 << PC_PM_NextLEDSet);
	}

	// **	Timer starten
	TCCR2 = (2 << CS20);
	TCCR1B = (1 << WGM12) | (3 << CS10);		// CTC Mode (OCR1A = Top) vorteiler 64
}

/********************************************************
*			PMode: Led Set Vorbereiten					*
********************************************************/

void PMode_PrepareLedSet() {
	ProgCtrlReg &= ~(1 << PC_PM_NextLEDSet);

	ProgPointer = (uint8_t*) &(ProgramDaten[NextEvent * 7]);
	NextEventTime = pgm_read_word(ProgPointer);  // Event Zeitpunkt laden

	// ** Derzeit inaktiven Helligkeits Cycle zuweisen (Backbuffer)
	unsigned char cyclePos = ((ModeOptions >> MC_MemPointer) & 0x01);
	
	//Helligkeitscycle alles aus
	for(i = 0; i < LED_HellCycleLeng; i++) {
		LED_Helligkeits_Cycle[cyclePos][i][0] = 0xFF;
		LED_Helligkeits_Cycle[cyclePos][i][1] = 0xFF;
		LED_Helligkeits_Cycle[cyclePos][i][2] = 0xFF;
		LED_Helligkeits_Cycle[cyclePos][i][3] = 0xFC;
	}
	
	// Starte Programmladen
	ProgPointer += 2;
	PMode_PrepareLedSetSC(CC_Red);

	ProgPointer += 4;
	PMode_PrepareLedSetSC(CC_Green);

	ProgPointer += 4;
	PMode_PrepareLedSetSC(CC_Blue);

	// Next Event erhhen
	NextEvent++;
	if (NextEvent == EventAnzahl) {
		NextEvent = 0;	// Es geht wieder mit Event 0 Los
	}
}


/********************************************************
*			PMode: Led Set Vorbereiten 1-8				*
********************************************************/

void PMode_PrepareLedSetSC(uint8_t farbe) {

	// LED 1
	tempProgDaten = pgm_read_byte(ProgPointer);
	tempHelligkeit = (tempProgDaten & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 0, tempHelligkeit);

	// LED 2
	tempHelligkeit = ((tempProgDaten >> 3) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 1, tempHelligkeit);

	// LED 3
	tempHelligkeit = ((tempProgDaten >> 6) & 0b00000011);
	tempProgDaten = pgm_read_byte(ProgPointer + 1);
	tempHelligkeit |= (tempProgDaten & 0b00000001) << 2;
	PMode_PrepareLedSetCC(farbe, 2, tempHelligkeit);

	// LED 4
	tempHelligkeit = ((tempProgDaten >> 1) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 3, tempHelligkeit);

	// LED 5
	tempHelligkeit = ((tempProgDaten >> 4) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 4, tempHelligkeit);

	// LED 6
	tempHelligkeit = ((tempProgDaten >> 7) & 0b00000001);
	tempProgDaten = pgm_read_byte(ProgPointer + 2);
	tempHelligkeit |= (tempProgDaten & 0b00000011) << 1;
	PMode_PrepareLedSetCC(farbe, 5, tempHelligkeit);

	// LED 7
	tempHelligkeit = ((tempProgDaten >> 2) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 6, tempHelligkeit);

	// LED 8
	tempHelligkeit = ((tempProgDaten >> 5) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 7, tempHelligkeit);

	// LED 9
	tempProgDaten = pgm_read_byte(ProgPointer + 3);
	tempHelligkeit = (tempProgDaten & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 8, tempHelligkeit);

	// LED 10
	tempHelligkeit = ((tempProgDaten >> 3) & 0b00000111);
	PMode_PrepareLedSetCC(farbe, 9, tempHelligkeit);
}

/********************************************************
*		PMode: Led Helligkeit in Cycle setzen			*
********************************************************/

void PMode_PrepareLedSetCC(uint8_t farbe, uint8_t LED, uint8_t hell) {
	unsigned char cyclePos = ((ModeOptions >> MC_MemPointer) & 0x01);
	if (LED < 8) {
		for(i = 0; i < hell; i++) {
			LED_Helligkeits_Cycle[cyclePos][i][farbe] &= ~(1 << LED);
		}
		
	}
	else {
		if (farbe == CC_Blue) {
			for(i = 0; i < hell; i++) {
				LED_Helligkeits_Cycle[cyclePos][i][CC_Rest] &= ~(1 << (LED - 4));
			}
		}
		else if (farbe == CC_Red) {
			for(i = 0; i < hell; i++) {
				LED_Helligkeits_Cycle[cyclePos][i][CC_Rest] &= ~(1 << (LED - 2));
			}
		}
		else {
			for(i = 0; i < hell; i++) {
				LED_Helligkeits_Cycle[cyclePos][i][CC_Rest] &= ~(1 << (LED - 6));
			}
		}
	}
}

/********************************************************
*			Direkten Modus Initialisieren				*
********************************************************/

void DMode_Init() {
}

/********************************************************
*			DMode: Led Set Vorbereiten					*
********************************************************/

void DMode_NewLedSet() {
}

/********************************************************
*********************************************************
*				Interupt Verarbeitung					*
*********************************************************
********************************************************/

/********************************************************
*				Timer Click VU Modus					*
********************************************************/

ISR(TIMER0_OVF_vect) {
	/* Interrupt Code */
}


/********************************************************
*			Helligkeitssteuerung der LEDs				*
********************************************************/

ISR(TIMER2_OVF_vect) {
	unsigned char cyclePos = (~((ModeOptions >> MC_MemPointer) & 0x01) & 0x01);

	PORTA = LED_Helligkeits_Cycle[cyclePos][LED_HellPosOffset][CC_Green];
	PORTB = LED_Helligkeits_Cycle[cyclePos][LED_HellPosOffset][CC_Red];
	PORTC = LED_Helligkeits_Cycle[cyclePos][LED_HellPosOffset][CC_Blue];
	PORTD = LED_Helligkeits_Cycle[cyclePos][LED_HellPosOffset][CC_Rest];

	LED_HellPosOffset++;
	if (LED_HellPosOffset == LED_HellCycleLeng) {
		LED_HellPosOffset = 0x00;
	}
}

/********************************************************
*	Nchster Programclick + auf Event Prfen			*
********************************************************/

ISR(TIMER1_COMPA_vect) {
	/*
	ProgTime++;
	PORTA = ~(1 << ProgTime);
	if(ProgTime == 8) {
		ProgTime = 0;
	}
	*/

	// ** Programm zeit Hochzhlen
	ProgTime++;

	// ** Prfen ob Programmende Erreicht
	if (ProgTime == ProgTime_Max) {
		ProgTime = 0x00;		// Zeit auf Null und weiter mit Prfen auf Event
	}
	if (ProgTime == NextEventTime) {
		// ** Aktuelle Event Helligkeit Cycle Tauschen und Befehl zum Event nachladen
		if ((ModeOptions & (1 << MC_MemPointer)) == 0) {
			ModeOptions |= (1 << MC_MemPointer);
		}
		else {
			ModeOptions &= ~(1 << MC_MemPointer);
		}
		ProgCtrlReg |= (1 << PC_PM_NextLEDSet);	// Befehl zum nachladen
	}
}

/********************************************************
*				Daten von PC empfangen					*
********************************************************/

ISR(USART_RXC_vect) {
	uint8_t CurByte = UDR;

	// falls bertragung luft
	if (CurCom == UCOM_DM_LEDSet) {
		
	}
	else if (CurCom == UCOM_DM_Flush) {

	}
	// befehl zum Moduswechseln
	else if (CurByte == UCOM_Goto_PMode) {
		ProgCtrlReg |= (1 << PC_Goto_PMode);
	}
	else if (CurByte == UCOM_Goto_DMode) {
		ProgCtrlReg |= (1 << PC_Goto_DMode);
	}
	else if (CurByte == UCOM_Goto_VUMode) {
		ProgCtrlReg |= (1 << PC_Goto_VUMode);
	}
	else if (CurByte == UCOM_Start_Bootload) {
		ProgCtrlReg |= (1 << PC_Start_Bootload);
	}
	// Direkt mode neues Set?
	else if (CurByte == UCOM_DM_LEDSet) {

	}
	else if (((ModeOptions >> MC_VUMode) & 0x01) == 0x01) {

	}
	else {
		// Unbekannter Befehl nachfragen und fertig
		UDR = UCOM_Return_Unknown;
	}
}



